#<profile4xi profile = "10" />
# -------------------------------------------------------------------------------------------------
#<program4xi program = "answer_challenge4xi.py">
#<abstract4xi abstract = "provides the function answer_challenge4xi" />
#<refer4xi reference = "connection.py" />
#<state4xi status = "ACK4tagging; ACK4moduletesting; NACK4systemtesting; [Rel 7.21]"/>
# -------------------------------------------------------------------------------------------------

# -------------------------------------------------------------------------------------------------
# Public API
# -------------------------------------------------------------------------------------------------
#<interface4xi interface = " 
__all__ = [ 'answer_challenge4xi' ]
#"
#/>

# -------------------------------------------------------------------------------------------------
# Standard Imports                          
# -------------------------------------------------------------------------------------------------
#<standardize4xi standard = "
import errno
import io
import itertools
import os
import pickle
import select
import socket
import struct
import sys
import time
import tempfile
try:
    import _winapi
    from _winapi import WAIT_OBJECT_0, WAIT_TIMEOUT, INFINITE
except ImportError:
    if sys.platform == 'win32':
        raise
    _winapi = None
    
import hmac     
#"
#/>

# -------------------------------------------------------------------------------------------------
# Multiprocessing Imports                          
# -------------------------------------------------------------------------------------------------
#<process4xi process = "
import _multiprocessing
from multiprocessing import connection
#"
#/>
 
# -------------------------------------------------------------------------------------------------
# DragonXi imports                              
# -------------------------------------------------------------------------------------------------
#<dragonize4xi dragonization = "
from constant_connection4xi import ConstantConnection4Xi 
from loop4xi import limit2loop4xi
from print4xi import print4xi
from time4xi import time4xi
from sleep4xi import omega2sleep4xi
#"
#/>

# -------------------------------------------------------------------------------------------------
#<function4xi function = "answer-challenge4xi" usage = "
#   connection.py
#       Listener:
#           __init__:
#               address = address or arbitrary_address(family) <-- forces an address !
#           accept:
#               c = self._listener.accept()
#               if self._authkey:
#                    deliver_challenge(c, self._authkey)
#                    answer_challenge(c, self._authkey)
#                    return c
#    client4xi(address, family=None, authkey=None):
#        if authkey is None:
#            authkey = os.urandom(32)
#        if authkey is not None:
#            try:
#                answer_challenge4xi(connection2xi, authkey)
#            ....                        
#            try:
#               deliver_challenge4xi(connection2xi, authkey)
#            ....            
#">
#<state4xi status = "ACK4review; ACK4tagging; NACK4moduletesting; NACK4systemtesting"/>
# -------------------------------------------------------------------------------------------------
def answer_challenge4xi(connection, authkey):
    '''
    answers challenge via given connection and with proper authkey
    phases: (1) connection.deliver_challenge(c, self.authkey) (2) connection.answer_challenge(c, self.authkey)
    '''
    import hmac                                          # Keyed-Hashing for Message Authentication

    
    # ---------------------------------------------------------------------------------------------    
    # (Step 1/8)
    # ---------------------------------------------------------------------------------------------        

    class2print = "answer_challenge4xi"
    method2print = "answer_challenge4xi"
    
    message2print = "{(Step 1/8a) started" + "; answer-challenge4xi-started: " + str(answer_challenge4xi)  + "; time: " + time4xi() + "}"
    print4xi(class2print, method2print, message2print)  
           
    message2print = "{(Step 1/8b) passed" + "; connection-passed-to-answer-challenge4xi: " + str(connection) + "; time: " + time4xi() + "}"
    print4xi(class2print, method2print, message2print) 
      
    message2print = "{(Step 1/8c) passed" + "; authkey-passed-to-answer-challenge4xi: " + str(authkey) + "; time: " + time4xi() + "}"
    print4xi(class2print, method2print, message2print) 
    
    
    # ---------------------------------------------------------------------------------------------    
    # (Step 2/8)
    # ---------------------------------------------------------------------------------------------        
    if authkey is None:
        authkey = os.urandom(32) #dragonized; create one if not given ! 
        
        message2print = "{(Step 2/8a) set" + "; authkey-set-by-answer-challenge4xi: " + str(authkey)  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)             
    if isinstance(authkey, bytes):
        
        message2print = "{(Step 2/8b) accepted" + "; authkey-accepted-by-answer-challenge4xi: " + "authkey" + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)              
    else:
        authkey = os.urandom(32) #dragonized; create proper one  !   
        
        message2print = "{(Step 2/8c) set" + "; authkey-set-by-answer-challenge4xi: " + str(authkey)  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print) 

    # ---------------------------------------------------------------------------------------------    
    # 
    # ---------------------------------------------------------------------------------------------                 
    cc2xi = ConstantConnection4Xi()
        
    message2print = "{(Step 2/8c) created" + "; cc2xi-created-by-answer-challenge4xi: " + str(cc2xi)  + "; time: " + time4xi() + "}"
    print4xi(class2print, method2print, message2print) 
               
    maximum2response = cc2xi.get_max_response_message_length()
    
    message2print = "{(Step 2/8d) set" + "; maximum2response-set-by-answer-challenge4xi to: " + str()  + "; time: " + time4xi() + "}"
    print4xi(class2print, method2print, message2print) 
       
    # ---------------------------------------------------------------------------------------------    
    # 
    # ---------------------------------------------------------------------------------------------                     
    challenge2xi = cc2xi.get_challenge()
    
    message2print = "{(Step 2/8e) set" + "; challenge2xi-set-by-answer-challenge4xi to: " + str(challenge2xi)  + "; time: " + time4xi() + "}"
    print4xi(class2print, method2print, message2print)    
                    
    message2print = "{(Step 2/8f) passing" + "; max-response-length-passing-to-connection-recv-bytes-by-answer-challenge4xi: " + str(maximum2response) + "; time: " + time4xi() + "}"
    print4xi(class2print, method2print, message2print)          
            
       
    # ---------------------------------------------------------------------------------------------    
    # 
    # ---------------------------------------------------------------------------------------------        
    ack2response = False
    nack2response = True
    
    max2count = limit2loop4xi(class2print, method2print) 
    message2print = "{(Step 2/8g1) got" + "; max2count-got-by-answer-challenge4xi: " + str(max2count)  + "; time: " + time4xi() + "}"
    print4xi(class2print, method2print, message2print)    
    
    count2try = 0        
    
    while ((ack2response is False) and (count2try < max2count)): 
    
        # -----------------------------------------------------------------------------------------    
        # dragonized - try several times to receive CHALLENGE-message while giving time to others
        # -----------------------------------------------------------------------------------------        
        ack2receive = False
        nack2receive = True
        try:
            # -------------------------------------------------------------------------------------    
            # receive anything
            # -------------------------------------------------------------------------------------         
            message2xi = connection.recv_bytes(maximum2response)         # dragonized

            message2print = "{(Step 2/8g2) received" + "; message2xi-received-by-answer-challenge4xi: " + str(message2xi)  + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print)              
            
            ack2receive = True
            nack2receive = False
            
            message2print = "{(Step 2/8g3) set" + "; ack2receive-set-by-answer-challenge4xi to: " + str(ack2receive)  + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print)    
        except:
            # -------------------------------------------------------------------------------------    
            # receiving failed
            # -------------------------------------------------------------------------------------            
            message2print = "{(Step 2/8g4) yielded" + "; nack2receive-yielded-by-answer-challenge4xi: " + str(nack2receive)  + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print) 
            
            count2try = count2try + 1 
            
        # -----------------------------------------------------------------------------------------    
        # 
        # -----------------------------------------------------------------------------------------
                     
        if ack2receive is True:
            # -------------------------------------------------------------------------------------    
            # make sure the received message is the right one
            # -------------------------------------------------------------------------------------                
            if message2xi[:len(challenge2xi)] == challenge2xi:           # header #CHALLENGE# tested
            
                # ---------------------------------------------------------------------------------    
                # CHALLENGE message received
                # ---------------------------------------------------------------------------------        

                ack2response = True
                nack2response = False
                
                message2xi = message2xi[len(challenge2xi):]              # filter the header away
    
                message2print = "{(Step 2/8h) built" + "; message2xi-built-by-answer-challenge4xi: " + str(message2xi)  + "; time: " + time4xi() + "}"
                print4xi(class2print, method2print, message2print)   
        
                            
            else:
            
                # ---------------------------------------------------------------------------------    
                # some other message received
                # ---------------------------------------------------------------------------------        
                message2print = "{warning" + "; message2xi-warning-by-answer-challenge4xi: " + " ** other smessage received: " + str(message2xi[:len(challenge2xi)])  + " ** " + "; time: " + time4xi() + "}"
                print4xi(class2print, method2print, message2print)                     
                                                                                 
                omega2sleep4xi(method2print, message2print)              # give time to others  
                count2try = count2try + 1                                # keep on trying
                 
        else:
            # -------------------------------------------------------------------------------------    
            # 
            # -------------------------------------------------------------------------------------           
            message2print = "{(Step 2/8g3) yielded" + "; nack2receive-yielded-by-answer-challenge4xi: " + str(nack2receive)  + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print) 
            
            omega2sleep4xi(method2print, message2print)                  # give time to others 
            count2try = count2try + 1
                         
        # -----------------------------------------------------------------------------------------    
        # end-of-while
        # -----------------------------------------------------------------------------------------        
           
    # ---------------------------------------------------------------------------------------------    
    # (Step 3/8)
    # ---------------------------------------------------------------------------------------------        
    if ack2response is False:
        message2print = "{(Step 3/8) yielded" + "; ack2response-yielded-by-answer-challenge4xi: " + str(ack2response)  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)     
    
    
        
    # ---------------------------------------------------------------------------------------------    
    # (Step 4/8)
    # ---------------------------------------------------------------------------------------------        
    ack2digest = False
    nack2digest = True
    
    if ack2response is True:
    
        # -----------------------------------------------------------------------------------------    
        # 
        # -----------------------------------------------------------------------------------------        
          
        message2print = "{(Step 4/8a) passing" + "; authkey-passing-to-hmac-new-by-answer-challenge4xi: " + str(authkey) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)          
                
        message2print = "{(Step 4/8b) passing" + "; message2xi-passing-to-hmac-new-by-answer-challenge4xi: " + str(message2xi) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)          
                  
        message2print = "{(Step 4/8c) calling" + "; hmac-new-calling-by-answer-challenge4xi: " + str(hmac.new) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)       
    
        # -----------------------------------------------------------------------------------------    
        # 
        # -----------------------------------------------------------------------------------------        
                  
        try: 
        
            digest2xi = hmac.new(authkey, message2xi).digest()   # create the digest of the strings passed
        
            message2print = "{(Step 4/8d1) created" + "; digest2xi-created-by-answer-challenge4xi: " + str(digest2xi)  + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print)         
            
            ack2digest = True
            nack2digest = False
            
            message2print = "{(Step 4/8d2) set" + "; ack2digest-set-by-answer-challenge4xi to: " + str(ack2digest)  + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print)    
            
            
        except:
            message2print = "{(Step 4/8e) yielded" + "; nack2digest-yielded-by-answer-challenge4xi: " + str(nack2digest)  + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print) 
    else:
    
        # -----------------------------------------------------------------------------------------    
        # 
        # -----------------------------------------------------------------------------------------        
         
        message2print = "{(Step 4/8f) yielded" + "; nack2ack2digest-yielded-by-answer-challenge4xi: " + str(nack2ack2digest)  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print) 
                
    # ---------------------------------------------------------------------------------------------    
    # (Step 5/8)
    # ---------------------------------------------------------------------------------------------        
    ack2send = False
    nack2send = True
    
    if ack2digest is True:
    
        # -----------------------------------------------------------------------------------------    
        # 
        # -----------------------------------------------------------------------------------------        
         
        message2print = "{(Step 5/8a) passing" + "; digest2xi-passing-by-answer-challenge4xi: " + str(digest2xi) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)          
       
        message2print = "{(Step 5/8b) calling" + "; function-calling-by-answer-challenge4xi: " + str(connection.send_bytes) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)                                    
    
        # -----------------------------------------------------------------------------------------    
        # 
        # -----------------------------------------------------------------------------------------        
                 
        try:    
            connection.send_bytes(digest2xi)
            
            message2print = "{(Step 5/8c1) sent" + "; digest2xi-sent-by-answer-challenge4xi: " + str(digest2xi)  + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print) 
       
            omega2sleep4xi(method2print, message2print)              # give time to others 
                 
            ack2send = True
            nack2send = False
            
            message2print = "{(Step 5/8c2) set" + "; ack2send-set-by-answer-challenge4xi to: " + str(ack2send)  + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print)    
    
        # -----------------------------------------------------------------------------------------    
        # 
        # -----------------------------------------------------------------------------------------        
  
        except:
            message2print = "{(Step 5/8d) yielded" + "; nack2send-yielded-by-answer-challenge4xi: " + str(nack2send)  + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print) 
    else:
        message2print = "{(Step 5/8e) yielded" + "; nack2send-yielded-by-answer-challenge4xi: " + str(nack2send)  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print) 
    
            
    # ---------------------------------------------------------------------------------------------    
    # (Step 6/8)
    # ---------------------------------------------------------------------------------------------        
    ack2response = False
    nack2response = True
    
    if ack2send is True:
    
        # -----------------------------------------------------------------------------------------    
        # 
        # -----------------------------------------------------------------------------------------        
         
        message2print = "{(Step 6/8a) passing" + "; max-response-message-length-passing-to-connection-recv-bytes-by-answer-challenge4xi: " + str(maximum2response) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)          
                
    
        message2print = "{(Step 6/8b) calling" + "; function-calling-by-answer-challenge4xi: " + str(connection.recv_bytes) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)                                    
        
          
        message2print = "{(Step 6/8c) warned" + "; message-length-warned-by-answer-challenge4xi: " + " ** large messages might be get rejected by message receiver: " + str(connection.recv_bytes) + " ** " + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)     
    
        # -----------------------------------------------------------------------------------------    
        # 
        # -----------------------------------------------------------------------------------------        
          
        try:
            omega2sleep4xi(method2print, message2print)                  # give time to others 
            
            response2xi = connection.recv_bytes(maximum2response)        # dragonized

            message2print = "{(Step 6/8d1) got" + "; response2xi-got-by-answer-challenge4xi: " + str(response2xi)  + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print)  
            
            ack2response = True
            nack2response = False
            
            message2print = "{(Step 6/8cd2) set" + "; ack2response-set-by-answer-challenge4xi to: " + str(ack2response)  + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print)    
                        
        except:
            message2print = "{(Step 6/8e) yielded" + "; nack2response-bytes-yielded-by-answer-challenge4xi: " + str(nack2response)  + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print) 
    else:
        message2print = "{(Step 6/8f) yielded" + "; nack2ack2response-yielded-by-answer-challenge4xi: " + str(nack2response)  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print) 
            
    # ---------------------------------------------------------------------------------------------    
    # (Step 7/8)
    # ---------------------------------------------------------------------------------------------        
    if ack2response is True:  
    
        # -----------------------------------------------------------------------------------------    
        # 
        # -----------------------------------------------------------------------------------------        
                 
        message2print = "{(Step 7/8a) testing" + "; response2xi-testing-by-answer-challenge4xi: " + str(response2xi)  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)      
        
        message2print = "{(Step 7/8b) checking" + "; welcome-checking-by-answer-challenge4xi: " + str(cc2xi.get_welcome()) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)   
        
        welcome2xi = cc2xi.get_welcome()
        
        if response2xi != welcome2xi:
    
            message2print = "{(Step 7/8c) yielded" + "; response2xi-yielded-by-answer-challenge4xi: " + str(response2xi)  + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print) 
        else:  
            
            message2print = "{(Step 7/8d) acknowledged" + "; response2xi-acknowledged-by-answer-challenge4xi: " + str(response2xi) + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print) 
    else:
    
        # -----------------------------------------------------------------------------------------    
        # 
        # -----------------------------------------------------------------------------------------        
         
        message2print = "{(Step 7/8e) yielded" + "; nack2response-yielded-by-answer-challenge4xi: " + str(nack2response)  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print) 
    
    # ---------------------------------------------------------------------------------------------    
    #(Step 8/8)  
    # ---------------------------------------------------------------------------------------------        
    message2print = "{(Step 8/8) returning" + "; answer-challenge4xi-returning: " + time4xi() + "}" 
    print4xi(class2print, method2print, message2print)  
      
    return
    
#</function4xi>
#</program4xi>


# -------------------------------------------------------------------------------------------------
# Recorded History                             
# ------------------------------------------------------------------------------------------------- 
#<record4xi record = "
#   test_none4xi.txt
#   sense-test-none4xi.xml
#   reason-test-none4xi.xml
#   ace-test-none4xi.xml
#"
#>

# -------------------------------------------------------------------------------------------------
# Legal Notice                             
# ------------------------------------------------------------------------------------------------- 
#<protect4xi protection = "copyright"
#   owner = "Small & Smart"
#   rights = "All rights reserved"
#   state = "California"
#   country = "United States of America"
#   references = "Legal notice at www.dragonxi.com; www.yourdragonxi.com; www.yourdragonxi.org; www.dragon2xi.com/; www.dragon4xi.com/; www.dragon8xi.com/ " 
#   codes = "Codes established in California and applicable to corporations in California"
#   compliance = "The California Law"
#   court = "Supreme Court of California, 350 McAllister Street, San Francisco, CA 94102-4783"
#/>
 
#<end4xi end = "
# -------------------------------------------------------------------------------------------------
# end of answer_challenge4xi.py                           
# ------------------------------------------------------------------------------------------------- 
#"
#/>


